#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>

#define SERVER_SOFTWARE		"nHTTPD alpha-20031020"

#define _DEF_SET_PORT		80
#define _DEF_SET_MAX_CLIENT	512
#define _DEF_SET_MAX_THREAD	SIGRTMAX - SIGRTMIN
#define _DEF_SET_LISTEN_BACKLOG	128
#define _DEF_SET_TIMEOUT	10
#define _DEF_SET_BUF_SIZE	32767
#define _DEF_SET_BUF_SIZE_MAX	1048576
#define _DEF_SET_FILE_CONFIG	"nhttpd.conf"
#define _DEF_SET_FILE_LOG	"nhttpd.log"
#define _DEF_SET_FILE_PID	"nhttpd.pid"
#define _DEF_SET_FILE_INDEX	"index.html"
#define _DEF_SET_DOCUMENT_ROOT	"htdocs"
#define _DEF_SET_CHARSET	"ISO-8859-1"
#define _DEF_SET_SERVER_USER	"nobody"
#define _DEF_ERROR_PAGE_400	"htdocs/error_400"
#define _DEF_ERROR_PAGE_401	"htdocs/error_401"
#define _DEF_ERROR_PAGE_403	"htdocs/error_403"
#define _DEF_ERROR_PAGE_404	"htdocs/error_404"

#define _DEF_CONTENT_TYPE	"text/plain"
#define _DEF_FILE_MIME		"mime_types.txt"

#define THREAD_STAT_NO_INIT	-1
#define THREAD_STAT_WAIT	0
#define THREAD_STAT_WORK	1

#define CLIENT_SLEEP		0
//#define CLIENT_INIT		1
#define CLIENT_REQUEST_READ	2
#define CLIENT_RESPONSE_HEAD	3
#define CLIENT_RESPONSE_WRITE	4
#define CLIENT_RESPONSE_BIG	5
//#define CLIENT_GC		6
//#define CLIENT_GC_WORK		7

#define PROTOCOL_PREPARSE_NO	0
#define PROTOCOL_PREPARSE_END	1
#define PROTOCOL_PREPARSE_ERROR	2

#define STAT_OK			200
#define STAT_NOT_MODIFIED	304
#define STAT_BAD_REQUEST	400
#define STAT_NOAUTH		401
#define STAT_FORBIDDEN		403
#define STAT_NOTFOUND		404

#define STATR_200		"OK"
#define STATR_304		"Not Modified"
#define STATR_400		"Bad Request"
#define STATR_401		"Unauthorized"
#define STATR_403		"Forbidden"
#define STATR_404		"Not Found"

#define HTTP_0_9		"HTTP/0.9"
#define HTTP_1_0		"HTTP/1.0"
#define HTTP_1_1		"HTTP/1.1"

#define METHOD_GET		"GET"
#define METHOD_HEAD		"HEAD"
#define METHOD_POST		"POST"

#define _CLIENT_DEF_BUF		255
#define _CLIENT_METHOD_BUF	4
#define _CLIENT_PROTOCOL_BUF	8
#define _CLIENT_FILENAME_BUF	2047
#define _CLIENT_URL_BUF		1023

#define REQUEST_REFERER		0x001
#define REQUEST_ENCODINGS	0x002
#define REQUEST_COOKIE		0x004
#define REQUEST_USERAGENT	0x008
#define REQUEST_ACCEPT		0x010
#define REQUEST_ACCEPTE		0x020

#define CGI_INTERFACE		"CGI/1.1"

//  Ȱ Ǵ Macro
#define NEW(t,n)        ((t*) malloc( sizeof(t) * (n) ))
#define MAX(a,b)        ((a) > (b) ? (a) : (b))
#define MIN(a,b)        ((a) < (b) ? (a) : (b))

//  ̴ ..

//ü   ʿ

typedef union {
	struct sockaddr sa;
	struct sockaddr_in sa_in;
} httpd_sockaddr;

typedef struct {
	char *domain;			//
	//unsigned short port;		//Ʈ   ʴ´
	char *document_root;		//ťƮƮ
	//char *file_log;		//α
} vhostinfo;

typedef struct{
	unsigned int port;			//Ʈ
	unsigned int max_client;		//ִ
	unsigned int max_thread;		//  ּ 1, ִ 32
	unsigned int listen_backlog;		//listen α
	unsigned int timeout;			//ŸӾƿ Ÿ̸

	unsigned long buf_size;			//⺻  
	unsigned long buf_size_max;		//⺻  ƽ 

	char *domain;				//⺻ 

	char *file_config;			// ̸
	char *file_log;				//α ̸
	char *file_pid;				//pid ̸
	char *file_index;			//ε ̸
	char *document_root;			//ťƮƮ
	char *default_charset;			//⺻  
	char *server_user;			// 

	char *error_page_400;			//
	char *error_page_401;
	char *error_page_403;
	char *error_page_404;

	unsigned int use_vhost;			//߾ ȣƮ   0̸  , 1 ̻ vhost  迭 ũ
	vhostinfo *vhost;			//߾ ȣƮ 

	unsigned short use_gzip;		//gzip   ,   
} config;

typedef struct {
	int pid;			//cgi pid

	char *server_software;		//Ʈ
	char *server_name;		// ȣƮ̸ IPּ
	char *gateway_interface;	//CGIŸ԰ 

	char *server_protocol;		//HTTP
	unsigned int server_port;	// Ʈ
	char *request_method;		//û method
	char *path_info;		//
	char *path_trans;		//
	char *script_name;		//Ǵ cgi URL
	char *query_string;		//?  URI

	char *remote_host;		//Ŭ̾Ʈ ȣƮ̸
	char *remote_addr;		//Ŭ̾Ʈ IP
	char *auth_type;		// 
	char *remote_user;		//Ŭ̾Ʈ  ̸( )
	char *remote_ident;		//RFC 931  Ȯ  쿡  ̸
	char *content_type;		//POST  û , ڰ Է  ¸ Ѵ. HTTP  νĵȴ
	unsigned int content_length;	//POST  ũ

	char *http_user_agent;		//Ŭ̾Ʈ 
	char *http_accept;		//
} cgiinfo;

typedef struct {
	int port;
	int listen_fd;
} httpd_server;

typedef struct {
	char *buf;			//û д 
	long read_size;			//û ü 
	long read_idx;			// û 
	unsigned short pre_parse;	//pre_parse
	unsigned short keep_alive;	//ŵ̺
	unsigned short static_file;	//Ϲ 
	time_t if_modified_since;	//if_modified_since
	char *method;			//û޽ (GET, HEAD)
	char *hostname;			//ȣƮ
	char *encodedurl;		//ûּ(ڵ )
	char *decodedurl;		//ڵּ
	char *origfilename;		//û̸
	char *query;			//GET Ʈ

	int vhost_no;			//0̸ ⺻
	int request_mask;		//

	char *referer;			//۷
	char *encodings;		//ڵ
	char *cookie;			//Ű
	char *useragent;		//Ʈ
	char *accept;			//
	char *accepte;			//
} http_request;

typedef struct {
	unsigned short status;		//ڵ (100~?00)
	unsigned short need_body;	//ٵ ʿ, HEAD ׻ 0
	struct stat filestat;		//ϻ
	int file;			//ϵũ
	unsigned long sended;		// Ʈ
} http_response;

typedef struct {
	short status;			//Ǯ (sleep , need_init , request_read д, response_head  , reponse_write , need_gc gc ó ٸ)
	httpd_sockaddr client_addr;	//Ŭ̾Ʈ  ü
	int conn_fd;			//
	char *protocol;			//
	short protocol_1_1;		//HTTP/1.1 ûΰ
	time_t last;			// û ð(̰  ©󳽴)
	http_request *request;
	http_response *response;
	unsigned short iscgi;		//cgi
	cgiinfo *cgi;
} client;

typedef struct {
	int pid;			// pid
	pthread_t *thread;		//
	pthread_mutex_t lock;		//ؽ
	int signum;			//Ҵ ñ׳ιȣ
	short status;			// (ٸ ̰ ִ)
	unsigned int clients;		//  Ŭ̾Ʈ 
	unsigned int client_c;		//ؾ  ִ Ŭ̾Ʈ 
	client *client_pool;
} thread;

typedef struct {
	unsigned int ext_len;		//Ȯ 
	char *ext;			//Ȯ
	char *content_type;		//ȭ Ÿ
	unsigned short static_content;	// (cgi )
} mime;

