#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <nhttpd.h>
#include <libhttpd.h>

void write_header( client *client_pool )
{
	char		*buf;
	int		buf_len;
	extern config	conf;
	char		*statr;
	char		*server_software = SERVER_SOFTWARE;
	extern char	*date_822;
	char		last_modified[50];
	struct tm	time_s;
	char		*content_type;


	buf = NEW( char, conf.buf_size + 8 );
	statr = status_code( client_pool->response->status );
	content_type = mime_type( client_pool->request );

	if ( client_pool->request->static_file == 1 )
	{
		buf_len = snprintf( buf, conf.buf_size, "%s %d %s\r\nDate: %s\r\nServer: %s\r\nContent-Type: %s\r\n", client_pool->protocol, client_pool->response->status, statr, (char *) (&date_822), server_software, content_type );
		if ( client_pool->request->if_modified_since == -1 )
		{
			gmtime_r( &(client_pool->response->filestat.st_mtime), &time_s );
			strftime( last_modified, 50, "%a, %d %b %G %k:%M:%S GMT", &time_s );
			buf_len += snprintf( &(buf[buf_len]), conf.buf_size, "Last-Modified: %s\r\nContent-Length: %d\r\n", last_modified, (int) client_pool->response->filestat.st_size );
		}
	}
	else
	{
		buf_len = snprintf( buf, conf.buf_size, "%s %d %s\r\nDate: %s\r\nServer: %s\r\n", client_pool->protocol, client_pool->response->status, statr, (char *) (&date_822), server_software );
	}

	if ( client_pool->request->keep_alive == 1 )
		buf_len += snprintf( &(buf[buf_len]), conf.buf_size, "Connection: Keep-alive\r\n" );
	else
		buf_len += snprintf( &(buf[buf_len]), conf.buf_size, "Connection: close\r\n" );

	buf = strcat( buf, "\r\n" );

	//printf("%s",buf);fflush(0);

	write( client_pool->conn_fd, buf, buf_len+2 );

	free( buf );

	return;
}

char *status_code( int status )
{
	switch ( status )
	{
		case STAT_OK :			return STATR_200; break;
		case STAT_NOT_MODIFIED :	return STATR_304; break;
		case STAT_BAD_REQUEST :		return STATR_400; break;
		case STAT_NOAUTH :		return STATR_401; break;
		case STAT_FORBIDDEN :		return STATR_403; break;
		case STAT_NOTFOUND :		return STATR_404; break;
	}

	return (char *) 0;
}

char *error_page_filename( int status )
{
	extern config	conf;
	switch ( status )
	{
		case STAT_BAD_REQUEST :	return conf.error_page_400; break;
		case STAT_NOAUTH :	return conf.error_page_401; break;
		case STAT_FORBIDDEN :	return conf.error_page_403; break;
		case STAT_NOTFOUND :	return conf.error_page_404; break;
	}

	return (char *) 0;
}

void httpd_set_err( client *client_pool, int error_code )
{
	client_pool->response->status = error_code;

	strncpy( client_pool->request->origfilename, error_page_filename( error_code ), _CLIENT_FILENAME_BUF );
	if ( strcmp( client_pool->protocol, HTTP_0_9 ) == 0 )
		client_pool->status = CLIENT_RESPONSE_WRITE;
	else
	{
		client_pool->status = CLIENT_RESPONSE_HEAD;
		stat( client_pool->request->origfilename, &(client_pool->response->filestat) );
	}

	stat( client_pool->request->origfilename, &(client_pool->response->filestat) );

	client_pool->request->keep_alive = 0;
	client_pool->response->need_body = 1;

	//printf("httpd_set_err: %d ", error_code);fflush(stdout);
	return;
}

void request_parse( http_request *request )
{
	char	*buf;
	char	*dummy;

	//printf("%s",request->buf);fflush(0);

	while ( ( buf = getbuf( request ) ) != (char*) 0 )
	{
		if ( buf[0] == '\0' )
			break;
		if ( strncasecmp( buf, "Referer:", 8 ) == 0 )
		{
			dummy = &buf[8];
			dummy += strspn( dummy, " \t" );
			strncpy( request->referer, dummy, _CLIENT_URL_BUF );
			request->request_mask = request->request_mask|REQUEST_REFERER;
		}
		else if ( strncasecmp( buf, "User-Agent:", 11 ) == 0 )
		{
			dummy = &buf[11];
			dummy += strspn( dummy, " \t" );
			strncpy( request->useragent, dummy, _CLIENT_DEF_BUF );
			request->request_mask = request->request_mask|REQUEST_USERAGENT;
		}
		else if ( strncasecmp( buf, "Host:", 5 ) == 0 )
		{
			dummy = &buf[5];
			dummy += strspn( dummy, " \t" );
			strncpy( request->hostname, dummy, _CLIENT_DEF_BUF );
			dummy = (char*) strchr( request->hostname, ':' );
			if ( dummy != (char*) 0 )
				*dummy = '\0';
		}
		else if ( strncasecmp( buf, "Accept:", 7 ) == 0 )
		{
			dummy = &buf[7];
			dummy += strspn( dummy, " \t" );
			strncpy( request->accept, dummy, _CLIENT_DEF_BUF );
			request->request_mask = request->request_mask|REQUEST_ACCEPT;
		}
		else if ( strncasecmp( buf, "Accept-Encoding:", 16 ) == 0 )
		{
			dummy = &buf[16];
			dummy += strspn( dummy, " \t" );
			strncpy( request->accepte, dummy, _CLIENT_DEF_BUF );
			request->request_mask = request->request_mask|REQUEST_ACCEPTE;
		}
		else if ( strncasecmp( buf, "If-Modified-Since:", 18 ) == 0 )
		{
			dummy = &buf[18];
			dummy += strspn( dummy, " \t" );
			request->if_modified_since = asctime2time_t( dummy );
		}
		else if ( strncasecmp( buf, "Cookie:", 7 ) == 0 )
		{
			dummy = &buf[7];
			dummy += strspn( dummy, " \t" );
			strncpy( request->cookie, dummy, _CLIENT_URL_BUF );
			request->request_mask = request->request_mask|REQUEST_COOKIE;
		}
		else if ( strncasecmp( buf, "Authorization:", 14 ) == 0 )
		{
			//dummy = &buf[14];
			//dummy += strspn( dummy, " \t" );
			//request->authorization = dummy;
		}
		else if ( strncasecmp( buf, "Content-Type:", 13 ) == 0 )
		{
			//dummy = &buf[13];
			//dummy += strspn( dummy, " \t" );
			//request->contenttype = (char*) atoi( dummy );
		}
		else if ( strncasecmp( buf, "Content-Length:", 15 ) == 0 )
		{
			//dummy = &buf[15];
			//request->contentlength = atol( dummy );
		}
		else if ( strncasecmp( buf, "Connection:", 11 ) == 0 )
		{
			dummy = &buf[11];
			dummy += strspn( dummy, " \t" );
			if ( strcasecmp( dummy, "keep-alive" ) == 0 )
				request->keep_alive = 1;
		}
	}

	return;
}

void request_pre_parse( client *client_pool )
{
	char	*request;
	char	*buf[3];
	int	idx;

	//printf("%s",client_pool->request->buf);fflush(0);

	for ( client_pool->request->read_idx = 0; client_pool->request->read_idx <= client_pool->request->read_size; ++client_pool->request->read_idx )
	{
		if ( client_pool->request->buf[client_pool->request->read_idx] == '\r' || client_pool->request->buf[client_pool->request->read_idx] == '\n' )
		{
			idx = client_pool->request->read_idx;
			if ( client_pool->request->buf[idx+1] == '\r' || client_pool->request->buf[idx+1] == '\n' )
				client_pool->request->read_idx+=2;

			request = NEW( char, ( idx + 1 ) );
			if ( request == (char *) 0 )
			{
				client_pool->request->pre_parse = PROTOCOL_PREPARSE_ERROR;
				return;
			}
			memset( request, 0x00, (idx + 1) );
			request = strncpy( request, (client_pool->request->buf), idx );
			break;
		}
	}

	if ( request == (char*) 0 )
	{
		strncpy( client_pool->protocol, HTTP_0_9, 8 );
		client_pool->request->pre_parse = PROTOCOL_PREPARSE_NO;
		free( request );
		return;
	}

	//printf("(%s)\n",request);fflush(0);
	buf[0] = request;
	buf[1] = (char *) strchr( buf[0], ' ' );
	if ( buf[1] == (char *) 0 )
	{
		strncpy( client_pool->protocol, HTTP_0_9, 8 );
		strncpy( client_pool->request->encodedurl, buf[0], _CLIENT_URL_BUF );
		client_pool->request->pre_parse = PROTOCOL_PREPARSE_END;
		free( request );
		return;
	}
	buf[1]++;
	buf[2] = (char *) strchr( buf[1], ' ') + 1;

	if ( strncasecmp( ( buf[0] ), METHOD_GET, 3 ) == 0 )
		strncpy( client_pool->request->method, METHOD_GET, 3 );
	else if ( strncasecmp( ( buf[0] ), METHOD_POST, 4 ) == 0 )
		strncpy( client_pool->request->method, METHOD_POST, 4 );
	else if ( strncasecmp( ( buf[0] ), METHOD_HEAD, 4 ) == 0 )
		strncpy( client_pool->request->method, METHOD_HEAD, 4 );
	else
	{
		client_pool->request->pre_parse = PROTOCOL_PREPARSE_ERROR;
		free( request );
		return;
	}

	if ( strncasecmp( ( buf[2] ), HTTP_1_0, _CLIENT_PROTOCOL_BUF ) == 0 )
		strncpy( client_pool->protocol, HTTP_1_0, _CLIENT_PROTOCOL_BUF );
	else if ( strncasecmp( ( buf[2] ), HTTP_1_1, _CLIENT_PROTOCOL_BUF ) == 0 )
	{
		strncpy( client_pool->protocol, HTTP_1_1, _CLIENT_PROTOCOL_BUF );
		client_pool->protocol_1_1 = 1;
	}
	else
	{
		client_pool->request->pre_parse = PROTOCOL_PREPARSE_ERROR;
		free( request );
		return;
	}

	client_pool->request->encodedurl = (char*) strncpy( client_pool->request->encodedurl, ( buf[1] ), ( strlen( buf[1] ) - _CLIENT_PROTOCOL_BUF - 1 ) );
	client_pool->request->pre_parse = PROTOCOL_PREPARSE_END;

	free( request );

	return;
}

void process_encodedurl( http_request *request )
{
	//vhost ó  
	extern config	conf;

	strdecode( request->decodedurl, request->encodedurl );
	de_dotdot( request->decodedurl );
	request->query = (char*) strchr( request->decodedurl, '?' );
	if ( request->query != (char *) 0 )
	{
		request->query[0] = '\0';
		request->query++;
	}

	if ( conf.use_vhost == 0 )
		snprintf( request->origfilename, _CLIENT_FILENAME_BUF, "%s%s", conf.document_root, request->decodedurl );
	//else
		//vhost ó

	if ( request->origfilename[( strlen( request->origfilename ) - 1 )] == '/' )
		request->origfilename = (char*) strcat( request->origfilename, conf.file_index );

	return;
}

char *getbuf( http_request *request )
{
	int	i;
	char	c;

	for ( i = request->read_idx; request->read_idx < request->read_size; ++request->read_idx )
	{
		c = request->buf[request->read_idx];
		if ( c == '\n' || c == '\r' )
		{
			request->buf[request->read_idx] = '\0';
			++request->read_idx;
			if ( c == '\r' && request->read_idx < request->read_size && request->buf[request->read_idx] == '\n' )
			{
				request->buf[request->read_idx] = '\0';
				++request->read_idx;
			}
			return &(request->buf[i]);
		}
	}
	return (char*) 0;
}

void strdecode( char *to, char *from )
{
	for ( ; *from != '\0'; ++to, ++from )
	{
		if ( from[0] == '%' && isxdigit( from[1] ) && isxdigit( from[2] ) )
		{
			*to = hexit( from[1] ) * 16 + hexit( from[2] );
			from += 2;
		}else{
			*to = *from;
		}
	}
	*to = '\0';
}


int hexit( char c )
{
	if ( c >= '0' && c <= '9' )
		return c - '0';
	if ( c >= 'a' && c <= 'f' )
		return c - 'a' + 10;
	if ( c >= 'A' && c <= 'F' )
		return c - 'A' + 10;
	return 0;
}

void de_dotdot( char *file )
{
	char	*cp;
	char	*cp2;
	int	l;

	while ( ( cp = (char *) strstr( file, "//") ) != (char *) 0 )
	{
		for ( cp2 = cp + 2; *cp2 == '/'; ++cp2 )
			continue;
		(void) strcpy( cp + 1, cp2 );
	}

	while ( ( cp = (char*) strstr( file, "/../" ) ) != (char*) 0 )
	{
		for ( cp2 = cp - 1; cp2 >= file && *cp2 != '/'; --cp2 )
			continue;
		if ( cp2 < file )
			break;
		(void) strcpy( cp2, cp + 3 );
	}

	while ( ( l = strlen( file ) ) > 3 && strcmp( ( cp = file + l - 3 ), "/.." ) == 0 )
	{
		for ( cp2 = cp - 1; cp2 >= file && *cp2 != '/'; --cp2 )
			continue;
		if ( cp2 < file )
			break;
		*cp2 = '\0';
	}
}

time_t asctime2time_t( char *asctime )
{
	int		asc_len;
	time_t		result = -1;
	struct tm	time;
	short		flag = 0;
	int		monthtab[12] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
	int		leap1, leap2, leapdays;

	asc_len = strlen( asctime );

	if ( asc_len <= 24 )
	{
		/*Sun Nov 6 08:49:30 1994         ; ANSI C's asctime() format*/
		time.tm_year =	atoi( &(asctime[( asc_len - 4 )]) );
		asctime[( asc_len - 5 )] =	'\0';
		time.tm_sec =	atoi( &(asctime[( asc_len - 7 )]) );
		time.tm_min =	atoi( &(asctime[( asc_len - 10 )]) );
		time.tm_hour =	atoi( &(asctime[( asc_len - 13 )]) );
		asctime[( asc_len - 14 )] =	'\0';
		asctime[7] =	'\0';
		time.tm_mon =	ascmon2int( &(asctime[4]) );
		time.tm_mday =	atoi( &(asctime[8]) );
		flag = 1;
	}
	else if ( asc_len == 29 )
	{
		/*Sun, 06 Nov 1994 08:49:30 GMT   ; RFC 822, updated by RFC 1123*/
		//IE  ַ   !
		if ( strcmp( &(asctime[( asc_len - 3 )]), "GMT" ) == 0 )
		{
			time.tm_mday =	atoi( &(asctime[5]) );
			time.tm_mon =	ascmon2int( &(asctime[8]) );
			asctime[16] = '\0';
			time.tm_year =	atoi( &(asctime[12]) );
			time.tm_hour =	atoi( &(asctime[17]) );
			time.tm_min =	atoi( &(asctime[20]) );
			time.tm_sec =	atoi( &(asctime[23]) );
			flag =1;
		}
	}
	else if ( asc_len < 36 )
	{
		/*Sunday, 06-Nov-94 08:49:30 GMT  ; RFC 850, obsoleted by RFC 1036*/
		if ( strcmp( &(asctime[( asc_len - 3 )]), "GMT" ) == 0 )
		{
			time.tm_sec =	atoi( &(asctime[( asc_len - 6 )]) );
			time.tm_min =	atoi( &(asctime[( asc_len - 9 )]) );
			time.tm_hour =	atoi( &(asctime[( asc_len - 12 )]) );
			asctime[( asc_len - 13 )] = '\0';
			time.tm_year =	atoi( &(asctime[( asc_len - 15 )]) );
			if ( time.tm_year > 70 )
				time.tm_year += 1900;
			else
				time.tm_year += 2000;
			time.tm_mon =	ascmon2int( &(asctime[( asc_len - 19 )]) );
			time.tm_mday =	atoi( &(asctime[( asc_len - 22 )]) );
			flag =1;
		}
	}

	if ( flag == 1 )
	{
		leap1 = ceil( 1970 / 4 );
		leap2 = floor( time.tm_year / 4 );
		leapdays = leap2 - leap1;
		for ( /*leap1*/; leap1 <= leap2; leap1+=4 )
		{
			if ( leap1/100 == 0 && leap1/400 != 0 )
				leapdays--;
		}
		result = ( ( ( time.tm_year - 1970 ) * 365 ) + monthtab[ ( time.tm_mon - 1 ) ] + time.tm_mday - 1 + leapdays ) * 24 * 60 * 60;
		result += time.tm_hour * 60 * 60;
		result += time.tm_min * 60;
		result += time.tm_sec;
	}

	//printf("(%d)",result);fflush(0);

	return result;
}

int ascmon2int( char *ascmonth )
{
	if ( strncmp( ascmonth, "Jan", 3 ) == 0 )	return 1;
	else if ( strncmp( ascmonth, "Feb", 3 ) == 0 )	return 2;
	else if ( strncmp( ascmonth, "Mar", 3 ) == 0 )	return 3;
	else if ( strncmp( ascmonth, "Apr", 3 ) == 0 )	return 4;
	else if ( strncmp( ascmonth, "May", 3 ) == 0 )	return 5;
	else if ( strncmp( ascmonth, "Jun", 3 ) == 0 )	return 6;
	else if ( strncmp( ascmonth, "Jul", 3 ) == 0 )	return 7;
	else if ( strncmp( ascmonth, "Aug", 3 ) == 0 )	return 8;
	else if ( strncmp( ascmonth, "Sep", 3 ) == 0 )	return 9;
	else if ( strncmp( ascmonth, "Oct", 3 ) == 0 )	return 10;
	else if ( strncmp( ascmonth, "Nov", 3 ) == 0 )	return 11;
	else if ( strncmp( ascmonth, "Dec", 3 ) == 0 )	return 12;
	return 0;
}

char *mime_type( http_request *request )
{
	char		*ext;
	int		len, i;
	extern int	mimes_len;
	extern mime	*mimes;

	if ( mimes_len <= 0 )
		return _DEF_CONTENT_TYPE;

	ext = rindex( request->origfilename, '.' );
	if ( ext == (char *) 0 )
	{
		return _DEF_CONTENT_TYPE;
	}
	ext++;
	len = strlen(ext);

	for ( i=0; i < mimes_len; i++ )
	{
		if ( mimes[i].ext_len == len && strncasecmp( mimes[i].ext, ext, len ) == 0 )
		{
			request->static_file = mimes[i].static_content;
			return mimes[i].content_type;
		}
	}

	return _DEF_CONTENT_TYPE;
}

mime *set_mime( mime *mimes, int *mimes_len )
{
	int		dummy;
	struct stat	filestat;
	char		*content_type;
	char		*static_content;
	int		statics;
	char		ext[1024];
	int		line_len, line_count, len;
	FILE		*fd;
	extern config	conf;
	int		charset_len;

	dummy = stat( _DEF_FILE_MIME, &filestat );
	if ( dummy < 0 || (int) filestat.st_size < 2 )
	{
		printf("%s file not found.\n",_DEF_FILE_MIME);
		return (mime *) 0;
	}

	fd = fopen( _DEF_FILE_MIME, "r" );
	if ( fd == NULL )
	{
		printf("%s file open error.\n",_DEF_FILE_MIME);
		return (mime *) 0;
	}

	line_count = 0;
	while ( fgets( ext, 1024, fd ) )
	{
		if ( ext[0] == '\n' || ext[0] == '\r' || ext[0] == '#' )
			continue;
		line_count++;
	}

	if ( line_count <= 0 )
		return (mime *) 0;

	mimes = NEW( mime, line_count );

	fseek( fd, 0, SEEK_SET );

	charset_len = strlen( conf.default_charset );

	line_count = 0;
        while ( fgets( ext, 1024, fd ) )
        {
		if ( ext[0] == '\n' || ext[0] == '\r' || ext[0] == '#' )
			continue;

		line_len = strlen( ext );
		if ( ext[line_len - 1] == '\r' || ext[line_len - 1] == '\n' )
		{
			ext[line_len - 1] = '\0';
			line_len--;
		}
		if ( ext[line_len - 1] == '\r' || ext[line_len - 1] == '\n' )
		{
			ext[line_len - 1] = '\0';
			line_len--;
		}

		content_type = index( ext, '\t' );
		if ( content_type == (char *) 0 )
			continue;
		content_type[0] = '\0';
		content_type++;
		content_type += strspn( content_type, " \t" );

		static_content = index( content_type, '\t' );
		if ( static_content == (char *) 0 )
			statics = 1;
		else
		{
			static_content[0] = '\0';
			static_content++;
			statics = atoi( static_content );
		}

		len = strlen( ext );
		mimes[line_count].ext_len = len;
		mimes[line_count].ext = NEW( char, len + 1 );
		strncpy( mimes[line_count].ext, ext, len);

		len = strlen( content_type );
		if ( strstr( content_type, "charset" ) != (char *) 0 )
		{
			mimes[line_count].content_type = NEW( char, len + charset_len );
			snprintf( mimes[line_count].content_type, ( len + charset_len ), content_type, conf.default_charset );
		}
		else
		{
			mimes[line_count].content_type = NEW( char, len + 1 );
			strncpy( mimes[line_count].content_type, content_type, len);
		}

		mimes[line_count].static_content = statics;

		line_count++;
	}

	fclose( fd );

	*mimes_len = line_count;

	return mimes;
}

int check_cgi( client *client_pool )
{
	int			filename_len, cgiext_len;
	extern config		conf;
	extern pthread_mutex_t	sender_mutex;
	extern pthread_cond_t	sender_cond;
	char			*dummy;

	cgiext_len=4;// Ͽ    ְ Ѵ.

	filename_len = strlen( client_pool->request->origfilename );

/*
	if ( filename_len > cgiext_len && strcasecmp( &(client_pool->request->origfilename[filename_len-cgiext_len]), ".cgi" ) == 0 )
	{
		client_pool->iscgi = 1;
		client_pool->cgi = NEW( cgiinfo, 1 );
		client_pool->response->need_body = 0;

		client_pool->cgi->server_software =		SERVER_SOFTWARE;
		client_pool->cgi->gateway_interface =		CGI_INTERFACE;
		client_pool->cgi->server_protocol =		client_pool->protocol;
		client_pool->cgi->request_method =		client_pool->request->method;
		client_pool->cgi->script_name =			client_pool->request->decodedurl;
		client_pool->cgi->query_string =		client_pool->request->query;
		client_pool->cgi->http_user_agent =		client_pool->request->useragent;
		client_pool->cgi->http_accept =			client_pool->request->accept;
		client_pool->cgi->remote_addr = NEW( char, 33 );
		dummy = inet_ntoa( client_pool->client_addr.sa_in.sin_addr );
		strncpy( client_pool->cgi->remote_addr, dummy, 32 );

		if ( conf.use_vhost == 0 )
		{
			client_pool->cgi->server_name =		conf.domain;
			client_pool->cgi->server_port =		conf.port;
		}
		else
		{
			//vhost ó
			client_pool->cgi->server_name =		conf.domain;
			client_pool->cgi->server_port =		conf.port;
		}

		//popen  ⼭ Ͼ.

		client_pool->status = CLIENT_RESPONSE_BIG;
		pthread_mutex_lock( &sender_mutex );
		pthread_cond_signal( &sender_cond );
		pthread_mutex_unlock( &sender_mutex );

		//client_pool->cgi->path_info
		//client_pool->cgi->path_trans
		//client_pool->cgi->remote_host
		//client_pool->cgi->auth_type
		//client_pool->cgi->remote_user
		//client_pool->cgi->remote_ident
		//client_pool->cgi->content_type
		//client_pool->cgi->content_length

	}
*/
	return client_pool->iscgi;
}
