#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/sendfile.h>

#include <nhttpd.h>
#include <libhttpd.h>

#define F_SETSIG	10

config		conf;
short   terminated;
mime		*mimes;
int		mimes_len;

httpd_server	*hs = (httpd_server *) 0;
thread		*threads = (thread *) 0;

pthread_mutex_t	sender_mutex;
pthread_cond_t	sender_cond;

time_t		now;
char		date_822[48];

void *proc( void *num );
void *sender( void *dummy );
void *listen_proc( void *dummy );
void client_first_init( client *client_pool );
void free_client_pool( client *client_pool );
void free_client_pool_data( client *client_pool );
httpd_server *httpd_sock_init( httpd_sockaddr *sa );
int initialize_socket( httpd_sockaddr *sa );
static void handle_term( int sig );
extern int get_conf( int argc, char **argv, config *conf );

int main( int argc, char **argv )
{
	httpd_sockaddr	sa;
	int		i, q;
	sigset_t	set;
	int		client_c;
	pthread_t	thread_sender;
	pthread_t	thread_listen;

	struct passwd	*pwd;
	uid_t		uid;
	gid_t		gid;

	struct tm	time_s;

	//
	if(get_conf( argc, argv, &conf ))
		pthread_exit(0);

	mimes = set_mime( mimes, &mimes_len );

	pwd = getpwnam( conf.server_user );
	if ( pwd == (struct passwd*) 0 )
	{
		(void) fprintf( stderr, "unknown user - '%s'\n", conf.server_user );
		exit( 1 );
	}
	uid = pwd->pw_uid;
	gid = pwd->pw_gid;

/*
	switch ( fork() )
	{
		case 0: break;
		case -1:
			printf("fork() failed");
			exit( 1 );
		default: exit( 0 );
	}
*/
	(void) setsid();
	//(void) fclose( stdin );
	//(void) fclose( stdout );
	//(void) fclose( stderr );

	(void) signal( SIGTERM, handle_term );
	(void) signal( SIGINT, handle_term );
	(void) signal( SIGPIPE, SIG_IGN );
	(void) signal( SIGHUP, handle_term );
	(void) signal( SIGUSR1, handle_term );

	// ʱȭ, 
	hs = httpd_sock_init( &sa );
	if ( hs == (httpd_server *) 0 )
		exit(1);

	//ü μ ñ׳ ũ 
	sigemptyset( &set );
	for ( i=SIGRTMIN; i <= SIGRTMAX; i++ )
		sigaddset( &set, i );

	pthread_sigmask( SIG_BLOCK, &set, NULL );

	//, Ŭ̾Ʈ Ǯ 
	time( &now );
	threads = NEW( thread, conf.max_thread );
	client_c = MAX( (int) ceil( (double) ( conf.max_client / conf.max_thread ) ), 1 );
	for ( i=0; i < conf.max_thread; i++ )
	{
		threads[i].pid = -1;
		threads[i].thread = NEW( pthread_t , 1 );
		threads[i].signum = SIGRTMAX - i;
		threads[i].status = THREAD_STAT_NO_INIT;
		threads[i].clients = 0;
		threads[i].client_c = client_c;
		threads[i].client_pool = NEW( client, client_c );

		for ( q=0; q < client_c; q++ )
		{
			client_first_init( &(threads[i].client_pool[q]) );
		}

		pthread_create( (pthread_t *) &(threads[i].thread), NULL, proc, (void *) &i );
		pthread_mutex_init( &(threads[i].lock), NULL );

		printf("%d Proc() thread start: Signal(%d)\n", ( i + 1 ), threads[i].signum);

		usleep(1);
	}

	//뷮   
	pthread_mutex_init( &sender_mutex, NULL );
	pthread_cond_init( &sender_cond, NULL );
	pthread_create( &thread_sender, NULL, sender, (void *) NULL );
	printf("sender() thread start\n");

	//listen 
	pthread_create( &thread_listen, NULL, listen_proc, (void *) NULL );
	printf("listen_proc() thread start\n");

	usleep(1);


	printf("nhttpd Started\n");fflush(stdout);


	for ( ; ; )
	{
		// ð Ѵ.  ۽ø ð  ߺؼ ϴ  
		time( &now );
		gmtime_r( &now, &time_s );
		strftime( date_822, 50, "%a, %d %b %G %k:%M:%S GMT", &time_s );
		sleep(1);

		if ( terminated == 1 )
			break;

		/*
		for ( i=0; i < conf.max_thread; i++ )
		{
			printf("%d ",threads[i].clients);
		}

		printf("\n");fflush(0);
		*/
		
	}


	pthread_exit(0);
}


void *proc( void *data )
{
	int		num = *( (int *) data );
	sigset_t	set;
	int		ret;
	int		i;
	struct siginfo	si;
	struct timespec	timer;
	int		request_read_res;
	long		send_res;

	threads[num].pid = getpid();

	sigemptyset( &set );
	sigaddset( &set, threads[num].signum );
	pthread_sigmask( SIG_BLOCK, &set, NULL );

	timer.tv_sec = conf.timeout;
	timer.tv_nsec = 0;

	for ( ; ; )
	{
		ret = sigtimedwait( &set, &si, &timer );

		//printf("SIGNAL!\n");fflush(0);
		//printf("(%d,%d,%d)\n",ret, si.si_code, si.si_fd);fflush(0);

		if ( terminated == 1 )
			break;

		if ( si.si_code == 0 && threads[num].client_c > 0 )
		{
			for ( i=0; i < threads[num].client_c; i++ )
			{
				if ( threads[num].client_pool[i].status == CLIENT_REQUEST_READ && ( threads[num].client_pool[i].last + 2 ) < now )
				{
					free_client_pool( &(threads[num].client_pool[i]) );
					pthread_mutex_lock( &(threads[num].lock) );
					threads[num].clients--;
					pthread_mutex_unlock( &(threads[num].lock) );
				}
			}
			pthread_mutex_lock( &(threads[num].lock) );
			if ( threads[num].clients < 0 )
				threads[num].clients = 0;
			pthread_mutex_unlock( &(threads[num].lock) );
			continue;
		}

/*
		if ( ret != threads[num].signum )
			continue;
*/

		for ( i=0; i < threads[num].client_c; i++ )
		{
			if ( threads[num].client_pool[i].status == CLIENT_REQUEST_READ )
			{
				if ( threads[num].client_pool[i].conn_fd == si.si_fd )
					break;
			}
		}

		if ( threads[num].client_pool[i].conn_fd != si.si_fd )
			continue;

		if ( si.si_code != POLL_IN || threads[num].client_pool[i].status != CLIENT_REQUEST_READ )
		{
			free_client_pool( &(threads[num].client_pool[i]) );
			pthread_mutex_lock( &(threads[num].lock) );
			threads[num].clients--;
			pthread_mutex_unlock( &(threads[num].lock) );
			continue;
		}

		if ( threads[num].client_pool[i].status == CLIENT_REQUEST_READ )
		{
			// ͸   ٸ ٽ ޸𸮸 ҴѴ
			if ( threads[num].client_pool[i].request->read_size > 0 && ( threads[num].client_pool[i].request->read_size ) % conf.buf_size == 0 )
				threads[num].client_pool[i].request->buf = realloc( threads[num].client_pool[i].request->buf, ( threads[num].client_pool[i].request->read_size + conf.buf_size ) );

			//Ͽ read
			request_read_res = read( threads[num].client_pool[i].conn_fd,
				&(threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size]), ( conf.buf_size - threads[num].client_pool[i].request->read_size ) );

			if ( request_read_res <= 0 )
			{
				free_client_pool( &(threads[num].client_pool[i]) );
				pthread_mutex_lock( &(threads[num].lock) );
				threads[num].clients--;
				pthread_mutex_unlock( &(threads[num].lock) );
				continue;
			}

			threads[num].client_pool[i].request->read_size += request_read_res;
			time( &( threads[num].client_pool[i].last ) );

			//request_pre_parse method(GET HEAD), û URL, HTTP  ȮѴ. HTTP/0.9  ϱ .
			//request_pre_parse ߸  ܰ ȴ. request_pre_parse   û   ǹ.
			if ( threads[num].client_pool[i].request->pre_parse != PROTOCOL_PREPARSE_END )
				request_pre_parse( &(threads[num].client_pool[i]) );

			if ( threads[num].client_pool[i].request->pre_parse == PROTOCOL_PREPARSE_NO )
			{
				//request_pre_parse  ʾҴ.
				continue;
			}
			else if ( threads[num].client_pool[i].request->pre_parse == PROTOCOL_PREPARSE_ERROR )
			{
				//request_pre_parse߿  ߻ û 
				httpd_set_err( &(threads[num].client_pool[i]), STAT_BAD_REQUEST );
			}
			else
			{
				if ( strcmp( (threads[num].client_pool[i].protocol), HTTP_0_9 ) == 0 )
				{
					//HTTP/0.9   븸 ϰ Ѵ.
					threads[num].client_pool[i].status = CLIENT_RESPONSE_WRITE;
				}
				else
				{
					//HTTP/0.9 ƴϹǷ û  ȮѴ.
					if ( ( threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size-1] == '\n' && threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size-2] == '\n' ) || ( threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size-1] == '\n' && threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size-2] == '\r' && threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size-3] == '\n' && threads[num].client_pool[i].request->buf[threads[num].client_pool[i].request->read_size-4] == '\r' ) )
					{
						request_parse( threads[num].client_pool[i].request );
						process_encodedurl( ( threads[num].client_pool[i].request ) );
						// vhost ó 

						if ( threads[num].client_pool[i].request->decodedurl[0] != '/' || ( threads[num].client_pool[i].request->decodedurl[1] == '.' && threads[num].client_pool[i].request->decodedurl[2] == '.' ) )
						{
							//û ϸ    û 
							httpd_set_err( &(threads[num].client_pool[i]), STAT_BAD_REQUEST );
						}
						else
						{
							//û Ͽ   ´.
							if ( stat( threads[num].client_pool[i].request->origfilename, &threads[num].client_pool[i].response->filestat ) != 0 )
							{
								//û  Ƿ 404 
								//printf("%s\n",threads[num].client_pool[i].request->origfilename);fflush(0);
								httpd_set_err( &(threads[num].client_pool[i]), STAT_NOTFOUND );
							}
							else
							{
								threads[num].client_pool[i].response->status = STAT_OK;

								if ( check_cgi( &(threads[num].client_pool[i]) ) == 1 )
								{
									//cgi
								}
								else
								{
									//û if_modified_since  ԵǾ ִٸ   ð   304  óѴ.
									if ( threads[num].client_pool[i].request->if_modified_since == threads[num].client_pool[i].response->filestat.st_mtime )
									{
										threads[num].client_pool[i].response->status = STAT_NOT_MODIFIED;
										threads[num].client_pool[i].response->need_body = 0;
									}

									//û ޽尡 HEAD ϶ body   ʴ´
									if ( strcmp( (threads[num].client_pool[i].request->method), METHOD_HEAD ) == 0 )
										threads[num].client_pool[i].response->need_body = 0;

									// ϰ Ѵ.
									threads[num].client_pool[i].status = CLIENT_RESPONSE_HEAD;
								}
							}
						}
					}
					else
						continue;	//û   ʾҴ.
				}

			}

		}

		if ( threads[num].client_pool[i].status == CLIENT_RESPONSE_HEAD )
		{
			// Ѵ.
			//printf("WRITE_HEAD ");fflush(0);
			write_header( &(threads[num].client_pool[i]) );

			//need_body  1̸ ٵ ϰ Ѵ.
			if ( threads[num].client_pool[i].response->need_body == 1 )
				threads[num].client_pool[i].status = CLIENT_RESPONSE_WRITE;
			else
			{
				// Ƿ keep-alive    û ٸ  Ѵ.
				if ( threads[num].client_pool[i].protocol_1_1 == 1 && threads[num].client_pool[i].request->keep_alive == 1 )
				{
					free_client_pool_data( &(threads[num].client_pool[i]) );
					threads[num].client_pool[i].status = CLIENT_REQUEST_READ;
				}
				else
				{
					free_client_pool( &(threads[num].client_pool[i]) );
					pthread_mutex_lock( &(threads[num].lock) );
					threads[num].clients--;
					pthread_mutex_unlock( &(threads[num].lock) );
					continue;
				}
			}
		}

		if ( threads[num].client_pool[i].status == CLIENT_RESPONSE_WRITE )
		{
			// Ѵ.
			//⼭ ū  츦 üũؼ ū ϸ ϴ  ѱ.
			//̰ ϸ ȵȴ -_-;;

			threads[num].client_pool[i].response->file = open( threads[num].client_pool[i].request->origfilename ,O_RDONLY );

			if ( (int) threads[num].client_pool[i].response->filestat.st_size > conf.buf_size )
			{
				threads[num].client_pool[i].status = CLIENT_RESPONSE_BIG;
				pthread_mutex_lock( &sender_mutex );
				pthread_cond_signal( &sender_cond );
				pthread_mutex_unlock( &sender_mutex );
				continue;
			}

			for ( ; ; )
			{
				//sendfile  ϰ  ū  ۿ 带   κ ؾ Ѵ.
				send_res = sendfile( threads[num].client_pool[i].conn_fd, threads[num].client_pool[i].response->file, &threads[num].client_pool[i].response->sended, (int) (threads[num].client_pool[i].response->filestat.st_size) );

				usleep(1);

				if ( send_res <= 0 )
				{
					threads[num].client_pool[i].request->keep_alive = 0;
					break;
				}

				if ( threads[num].client_pool[i].response->sended >= (int) threads[num].client_pool[i].response->filestat.st_size )
					break;
			}

			//printf("SEND_END");fflush(0);
			// Ƿ keep-alive    û ٸ  Ѵ.
			if ( threads[num].client_pool[i].protocol_1_1 == 1 && threads[num].client_pool[i].request->keep_alive == 1 )
			{
				free_client_pool_data( &(threads[num].client_pool[i]) );
				threads[num].client_pool[i].status = CLIENT_REQUEST_READ;
			}
			else
			{
				free_client_pool( &(threads[num].client_pool[i]) );
				pthread_mutex_lock( &(threads[num].lock) );
				threads[num].clients--;
				pthread_mutex_unlock( &(threads[num].lock) );
				continue;
			}
		}
	}

	pthread_exit(0);
}


void *sender( void *dummy )
{
	int		i, p;
	int		send_res;
	int		sending = 0;

	while( 1 )
	{
		if ( sending == 0 )
		{
			pthread_mutex_lock( &sender_mutex );
			pthread_cond_wait( &sender_cond, &sender_mutex );
			pthread_mutex_unlock( &sender_mutex );
		}

		if ( terminated == 1 )
			break;

		sending = 0;

		for ( i=0; i < conf.max_thread; i++ )
		{
			if ( threads[i].clients > 0 ) //ó Ŭ̾Ʈ ڰ ̻  Ƚϰ Ǭ.
			{
				for ( p=0; p < threads[i].clients ; p++ )
				{
					if ( threads[i].client_pool[p].status == CLIENT_RESPONSE_BIG )
					{
						sending++;
						if ( threads[i].client_pool[p].iscgi == 1 )
						{
							usleep(1);
							//cgi  Ѹ Ǵ cgi   带 .
						}
						else
						{
							//printf("%d: %s\n",i, threads[i].client_pool[p].request->origfilename);fflush(0);
							send_res = sendfile( threads[i].client_pool[p].conn_fd, threads[i].client_pool[p].response->file, &( threads[i].client_pool[p].response->sended ), conf.buf_size );

							usleep(1);

							if ( send_res <= 0 )
							{
								sending--;
								free_client_pool( &(threads[i].client_pool[p]) );
								pthread_mutex_lock( &(threads[i].lock) );
								threads[i].clients--;
								pthread_mutex_unlock( &(threads[i].lock) );
								continue;
							}

							if ( threads[i].client_pool[p].response->sended >= (int) threads[i].client_pool[p].response->filestat.st_size )
							{
								sending--;
								if ( threads[i].client_pool[p].protocol_1_1 == 1 && threads[i].client_pool[p].request->keep_alive == 1 )
								{
									free_client_pool_data( &(threads[i].client_pool[p]) );
									threads[i].client_pool[p].status = CLIENT_REQUEST_READ;
								}
								else
								{
									free_client_pool( &(threads[i].client_pool[p]) );
									pthread_mutex_lock( &(threads[i].lock) );
									threads[i].clients--;
									pthread_mutex_unlock( &(threads[i].lock) );
								}
							}
						}
					}
				}
			}
		}
	}

	pthread_exit(0);
}

void *listen_proc( void *dummy )
{
	struct pollfd	listen_poll;
	int		poll_res;
	int		accept_fd;
	int		i;
	int		sz;
	int		min_thread, min_clients;
	int		on;
	httpd_sockaddr	sa;

	listen_poll.fd = hs->listen_fd;
	listen_poll.events = POLLIN;

	on = 1;

	while ( 1 )
	{
		poll_res = poll( &listen_poll, 1, 1000 );

		if ( terminated == 1 )
			break;

		if ( poll_res == 0 )
			continue;

		sz = sizeof( sa );
		accept_fd = accept( hs->listen_fd, &sa.sa, &sz );

		if ( fcntl( accept_fd, F_SETFL, O_RDWR|O_NONBLOCK|O_ASYNC ) < 0 )
		{
		}

		(void) setsockopt( accept_fd, SOL_SOCKET, SO_LINGER, (char*) &on, sizeof(on) );

		//   ϰ ִ 带 ã´.
		min_thread = -1;
		min_clients = conf.max_client;
		for ( i=0; i < conf.max_thread; i++ )
		{
			if ( threads[i].clients < min_clients )
			{
				min_thread = i;
				min_clients = threads[i].clients;
			}
		}

		//ã  client_pool ϰ RTS ҴѴ
		for ( i=0; i < threads[min_thread].client_c; i++ )
		{
			if ( threads[min_thread].client_pool[i].status == CLIENT_SLEEP )
			{
				memset( &(threads[min_thread].client_pool[i].client_addr), 0, sizeof(threads[min_thread].client_pool[i].client_addr) );
				memcpy( &(threads[min_thread].client_pool[i].client_addr), &sa, sizeof(struct sockaddr_in) );

				threads[min_thread].client_pool[i].conn_fd = accept_fd;
				threads[min_thread].client_pool[i].status = CLIENT_REQUEST_READ;

				if ( fcntl( accept_fd, F_SETSIG, threads[min_thread].signum ) < 0 )
				{
				}

				if ( fcntl( accept_fd, F_SETOWN, threads[min_thread].pid ) < 0 )
				{
				}

				pthread_mutex_lock( &(threads[min_thread].lock) );
				threads[min_thread].clients++;
				pthread_mutex_unlock( &(threads[min_thread].lock) );

				time( &( threads[min_thread].client_pool[i].last) );

				break;
			}
		}

		//pthread_kill( threads[min_thread].thread, threads[min_thread].signum );
	}

	pthread_exit(0);
}

void client_first_init( client *client_pool )
{
	client_pool->status =			CLIENT_SLEEP;
	client_pool->conn_fd =			-1;
	//client_pool->client_addr = NEW( httpd_sockaddr, 1 );
	client_pool->protocol =			NEW( char, _CLIENT_PROTOCOL_BUF + 1 );
	client_pool->protocol_1_1 =			0;
	client_pool->request =			NEW( http_request, 1 );
	client_pool->response =			NEW( http_response, 1 );
	client_pool->iscgi =			0;

	client_pool->request->buf =		NEW( char, conf.buf_size + 1 );
	client_pool->request->method =		NEW( char, _CLIENT_METHOD_BUF + 1 );
	client_pool->request->hostname =	NEW( char, _CLIENT_DEF_BUF + 1 );
	client_pool->request->encodedurl =	NEW( char, _CLIENT_URL_BUF + 1 );
	client_pool->request->decodedurl =	NEW( char, ( _CLIENT_URL_BUF * 2 ) + 1 );
	client_pool->request->origfilename =	NEW( char, _CLIENT_FILENAME_BUF + 1 );
	//client_pool->request->query =		NEW( char, _CLIENT_URL_BUF + 1 );
	client_pool->request->referer =		NEW( char, _CLIENT_URL_BUF + 1 );
	client_pool->request->encodings =	NEW( char, _CLIENT_DEF_BUF + 1 );
	client_pool->request->cookie =		NEW( char, _CLIENT_URL_BUF + 1 );
	client_pool->request->useragent =	NEW( char, _CLIENT_DEF_BUF + 1 );
	client_pool->request->accept =		NEW( char, _CLIENT_DEF_BUF + 1 );
	client_pool->request->accepte =		NEW( char, _CLIENT_DEF_BUF + 1 );

	client_pool->request->if_modified_since =	-1;
	client_pool->request->read_size =		0;
	client_pool->request->pre_parse =		PROTOCOL_PREPARSE_NO;
	client_pool->request->read_idx =		0;
	client_pool->request->keep_alive =		0;
	client_pool->request->static_file =		1;
	client_pool->request->request_mask =		0;
	client_pool->request->vhost_no =		0;

	client_pool->response->status =			0;
	client_pool->response->need_body =		1;

	client_pool->response->file =			-1;
	client_pool->response->need_body =		1;
	client_pool->response->sended =			0;
}

void free_client_pool( client *client_pool )
{
	free_client_pool_data( client_pool );

	memset( &( client_pool->client_addr ), 0x00, sizeof( httpd_sockaddr ) );

	if ( client_pool->conn_fd != -1 )
	{
		close( client_pool->conn_fd );
		client_pool->conn_fd =			-1;
	}

	client_pool->status =				CLIENT_SLEEP;

	return;
}

void free_client_pool_data( client *client_pool )
{
	memset( client_pool->protocol,			0x00, _CLIENT_PROTOCOL_BUF + 1 );
	client_pool->protocol_1_1 =			0;
	if ( client_pool->iscgi == 1 )
	{
		free( client_pool->cgi->remote_addr );
		free( client_pool->cgi );
		client_pool->iscgi =			0;
		//cgi ̴ ۾ ־ Ѵ.
	}

	memset( client_pool->request->method,		0x00, _CLIENT_METHOD_BUF + 1 );
	memset( client_pool->request->hostname,		0x00, _CLIENT_DEF_BUF + 1 );
	memset( client_pool->request->encodedurl,	0x00, _CLIENT_URL_BUF + 1 );
	memset( client_pool->request->decodedurl,	0x00, ( _CLIENT_URL_BUF * 2 ) + 1 );
	memset( client_pool->request->origfilename,	0x00, _CLIENT_FILENAME_BUF + 1 );
	//memset( client_pool->request->query,		0x00, _CLIENT_URL_BUF + 1 );

	if ( client_pool->request->request_mask & REQUEST_REFERER )
		memset( client_pool->request->referer,		0x00, _CLIENT_URL_BUF + 1 );
	if ( client_pool->request->request_mask & REQUEST_ENCODINGS )
		memset( client_pool->request->encodings,	0x00, _CLIENT_DEF_BUF + 1 );
	if ( client_pool->request->request_mask & REQUEST_COOKIE )
		memset( client_pool->request->cookie,		0x00, _CLIENT_URL_BUF + 1 );
	if ( client_pool->request->request_mask & REQUEST_USERAGENT )
		memset( client_pool->request->useragent,	0x00, _CLIENT_DEF_BUF + 1 );
	if ( client_pool->request->request_mask & REQUEST_ACCEPT )
		memset( client_pool->request->accept,		0x00, _CLIENT_DEF_BUF + 1 );
	if ( client_pool->request->request_mask & REQUEST_ACCEPTE )
		memset( client_pool->request->accepte,		0x00, _CLIENT_DEF_BUF + 1 );

	if ( client_pool->request->read_size > conf.buf_size )
	{
		free( client_pool->request->buf );
		client_pool->request->buf =		NEW( char, conf.buf_size + 1 );
	}
	memset( client_pool->request->buf,		0x00, conf.buf_size + 1 );

	client_pool->request->if_modified_since =	-1;
	client_pool->request->read_size =		0;
	client_pool->request->pre_parse =		PROTOCOL_PREPARSE_NO;
	client_pool->request->read_idx =		0;
	client_pool->request->keep_alive =		0;
	client_pool->request->static_file =		1;
	client_pool->request->request_mask =		0;
	client_pool->request->vhost_no =		0;

	client_pool->response->status =			0;
	client_pool->response->need_body =		1;
	client_pool->response->sended =			0;

	if ( client_pool->response->file != -1 )
	{
		close( client_pool->response->file );
		client_pool->response->file =		-1;
	}

}

httpd_server *httpd_sock_init( httpd_sockaddr *sa )
{
	httpd_server	*hs;

	hs = NEW( httpd_server, 1 );
	if ( hs == (httpd_server *) 0 )
		return (httpd_server *) 0;

	hs->port = conf.port;

	if ( sa == (httpd_sockaddr*) 0 )
		hs->listen_fd = -1;
	else
	{
		sa->sa.sa_family = AF_INET;
		sa->sa_in.sin_addr.s_addr = htonl( INADDR_ANY );
		sa->sa_in.sin_port = htons( conf.port );
		hs->listen_fd = initialize_socket( sa );
	}

	if ( hs->listen_fd == -1)
	{
		free( hs );
		return (httpd_server*) 0;
	}

	return hs;
}

int initialize_socket( httpd_sockaddr *sa )
{
	int	listen_fd;
        int	on, flags;

        listen_fd = socket( sa->sa.sa_family, SOCK_STREAM, 0 );

        if ( listen_fd < 0 )
                return -1;

	(void) fcntl( listen_fd, F_SETFD, 1 );
        (void) fcntl( listen_fd, F_SETFL, O_NONBLOCK);
        on = 1;
        (void) setsockopt( listen_fd, SOL_SOCKET, SO_REUSEADDR, (char*) &on, sizeof(on) );
        if ( bind( listen_fd, &sa->sa, sizeof(struct sockaddr_in) ) < 0 )
        {
                (void) close( listen_fd );
                return -1;
        }

        flags = fcntl( listen_fd, F_GETFL, 0 );
        if ( flags == -1 )
        {
                (void) close( listen_fd );
                return -1;
        }

        if ( fcntl( listen_fd, F_SETFL, flags | O_NDELAY ) < 0 )
        {
                (void) close( listen_fd );
                return -1;
        }

	if ( listen( listen_fd, conf.listen_backlog ) < 0 )
        {
                (void) close( listen_fd );
                return -1;
        }

        return listen_fd;
}

static void handle_term( int sig )
{
	int	i;

	if ( terminated == 1 )
		return;

	terminated = 1;
	close( hs->listen_fd );

	pthread_mutex_lock( &sender_mutex );
	pthread_cond_signal( &sender_cond );
	pthread_mutex_unlock( &sender_mutex );

	for ( i=0; i < conf.max_thread; i++ )
	{
		kill( threads[i].pid, threads[i].signum );
		printf("%d ", ( i + 1 ));
	}
	printf("thread killed.\n");fflush(stdout);
	exit( 0 );
}

